#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from __future__ import with_statement, division

"""Read a gpx file from a GPS device, and produce subtitles with the speed at a given moment. 

Speed Subtitles
Copyright © 2008 Maarten Sneep. 

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__author__ = "Maarten Sneep (maarten.sneep@xs4all.nl)"
__version__ = "0.1"
__date__ = "14-08-2008"
__copyright__ = "Copyright © 2008 Maarten Sneep"
__license__ = "GPL version 2 or later."


import os
import sys
import re
import random
import tempfile
import zipfile
from datetime import datetime, timedelta
import math
import re
import codecs
import urllib
from xml.dom import minidom
from optparse import OptionParser

def Main():
    (options, args) = cmdline()
    
    if len(args) == 0:
        infile = sys.stdin
        outfile = sys.stdout
    elif len(args) == 1:
        infile = urllib.urlopen(args[0])
        outfile = sys.stdout
    else:
        infile = urllib.urlopen(args[0])
        outfile = codecs.open(args[1], "w", "iso-8859-1")
    
    gpx = minidom.parse(infile)
    
    (time, speed, duration) = handleGPX(gpx, options.start, options.end, options.duration)
    
    makeQTtext(outfile, time, speed, options, duration)
    
    infile.close()
    outfile.close()
    sys.exit(0)

def distance(loc1, loc2):
    """Calculate geodesic distance (in m) between two points specified by
latitude/longitude (in numeric degrees) using Vincenty inverse formula 
for ellipsoids. 

Arguments are tuples of (lat, lon) pairs. 

Code based on a Javascript version by © 2002-2008 Chris Veness. 

That code contains the following: "You are welcome to re-use these scripts 
[without any warranty express or implied] provided you retain my copyright notice 
and when possible a link to my website (under a LGPL license)."

The originating web page: http://www.movable-type.co.uk/scripts/latlong-vincenty.html
"""
    a = 6378137 # in m
    b = 6356752.3142 # in m
    f = 1/298.257223563
    deg2rad = math.pi / 180
    lat1, lon1 = (loc1[0] * deg2rad, loc1[1] * deg2rad)
    lat2, lon2 = (loc2[0] * deg2rad, loc2[1] * deg2rad)
    L = (lon2 - lon1)
    U1 = math.atan((1-f) * math.tan(lat1))
    U2 = math.atan((1-f) * math.tan(lat2))
    sinU1 = math.sin(U1)
    cosU1 = math.cos(U1)
    sinU2 = math.sin(U2)
    cosU2 = math.cos(U2)
    Lambda = L
    LambdaP = 2*math.pi
    iterLimit = 20
    
    while math.fabs(Lambda-LambdaP) > 1e-12 and iterLimit > 0:
        sinLambda = math.sin(Lambda)
        cosLambda = math.cos(Lambda)
        sinSigma = math.sqrt((cosU2*sinLambda)**2 + (cosU1*sinU2-sinU1*cosU2*cosLambda)**2)
        if (sinSigma==0): return 0 # co-incident points
        cosSigma = sinU1*sinU2 + cosU1*cosU2*cosLambda
        sigma = math.atan2(sinSigma, cosSigma)
        # sinSigma ≠ 0, already caught by condition above.
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha**2
        try:
            cos2SigmaM = cosSigma - 2*sinU1*sinU2/cosSqAlpha
        except:
            cos2SigmaM = 0 # equatorial line: cosSqAlpha=0
        C = f/16*cosSqAlpha*(4+f*(4-3*cosSqAlpha));
        LambdaP = Lambda;
        Lambda = L + (1-C) * f * sinAlpha * (sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*cos2SigmaM**2)))
        iterLimit -= 1
    
    if iterLimit == 0:
        return None  # formula failed to converge
    
    uSq = cosSqAlpha * (a**2 - b**2) / (b**2)
    A = 1 + uSq/16384*(4096+uSq*(-768+uSq*(320-175*uSq)))
    B = uSq/1024 * (256+uSq*(-128+uSq*(74-47*uSq)))
    deltaSigma = B*sinSigma*(cos2SigmaM+B/4*(cosSigma*(-1+2*cos2SigmaM**2) - B/6*cos2SigmaM*(-3+4*sinSigma**2)*(-3+4*cos2SigmaM**2)))
    return b*A*(sigma-deltaSigma);

def parseTimeString(str):
    if str is None: 
        return None
        
    list = [int(val) for val in re.split("-|T|:|Z", str)[0:-1]]
    year, month, day, hour, minute, second = tuple(list)
    return datetime(year, month, day, hour, minute, second, 0, None)

def handleGPX(gpx, startstr, endstr, durationstr):
    if (startstr is None and endstr is None) or durationstr is None: 
        sys.stderr.write("Could not determine the recodring time")
        sys.exit(1)
    
    start = parseTimeString(startstr)
    end = parseTimeString(endstr)
    duration = durationstr.split(":")
    if len(duration) == 3:
        durationdelta = timedelta(0, int(duration[2]), 0, 0, 
            int(duration[1]), int(duration[0]))
    elif len(duration) == 2:
        durationdelta = timedelta(0, int(duration[1]), 0, 0, 
            int(duration[0]))
    else:
        f, i = math.modf(float(duration[0]))
        durationdelta = timedelta(0, int(i), 0, int(1000*f))
    
    if start is None:
        start = end - durationdelta
    else:
        end = start + durationdelta
    
    times = []
    latitudes = []
    longitudes = []
    
    points = gpx.getElementsByTagName("trkpt")
    for point in points:
        time = parseTimeString(point.getElementsByTagName("time")[0].firstChild.data)
        times.append(time-start)
        latitudes.append(float(point.getAttribute("lat")))
        longitudes.append(float(point.getAttribute("lon")))
    
    idx = 0 
    while times[idx] < timedelta(0): idx += 1
    if idx > 1: 
        times = times[idx-1:]
        latitudes = latitudes[idx-1:]
        longitudes = longitudes[idx-1:]
    
    idx = -1
    while times[idx] > durationdelta: idx -= 1
    if idx < -1: 
        times = times[:idx+1]
        latitudes = latitudes[:idx+1]
        longitudes = longitudes[:idx+1]
    
    speeds = []
    for idx in range(len(times)-1):
        delta = times[idx+1] - times[idx]
        d = distance( (latitudes[idx], longitudes[idx]), 
            (latitudes[idx+1], longitudes[idx+1]) )
        t = (float(delta.seconds) + delta.microseconds*1.0e-6)
        speed = 3.6 * (d / t)
        speeds.append(speed)
        if speed > 40 or (idx > 2 and speeds[-2] > 40):
            sys.stdout.write("(%f, %f) -> (%f, %f) [%d]: %f/%f = %f\n" % 
            (latitudes[idx], longitudes[idx], latitudes[idx+1], longitudes[idx+1], idx, d, t, speed))
    speeds.append(0)
    
    return (times, speeds, durationdelta)
    
def makeOffsetString(delta):
    # [00:00:45.00]
    seconds = delta.seconds % 60
    minutes = (delta.seconds // 60) % 60
    hours = delta.seconds // 3600
    
    return "[%02d:%02d:%02d.00]\n" % (hours, minutes, seconds)

def makeQTtext(outfile, times, speeds, options, duration):
    size = options.size.split(",")
    for idx in range(len(size)): size[idx] = int(size[idx])
    if size[1] < 45: size[1] = 45
    if size[0] < 120: size[0] = 120
    fontsize = int(0.08*size[1])
    if fontsize < 14: fontsize = 14
    
    location = list(options.location.upper())
    if 'L' in location: 
        alignment = "left"
    elif 'R' in location:
        alignment = "right"
    else:
        alignment = "center"
    
    box = "%d, %d, %d, %d" % (20, 20, size[1]-20, size[0]-20)
    
    if 'T' in location:
        textPrefix = ""
    elif 'B' in location:
        textPrefix = "\n\n\n\n\n\n\n\n\n\n"
    else:
        textPrefix = "\n\n\n\n\n"
    
    header = """{QTtext}
{font:Helvetica}
{bold}{size:%(fontsize)d}
{timeScale:100}
{width:%(width)d}
{height:%(height)d}
{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}
{timeStamps:absolute}
{language:0}{textEncoding:0}
{justify:%(alignment)s}
{anti-alias:on}
{textBox: %(textBox)s}
{doNotAutoScale: on}
{clipToTextbox: on}
{keyedText:on}
{dropShadow:on}{dropShadowOffset:3,3}{dropShadowTransparency:200}
[00:00:00.00]
""" % { 'width': size[0], 
        'height': size[1], 
        'fontsize': fontsize, 
        'textBox': box,
        'alignment': alignment}
    
    outfile.write(header)
    
    startidx = 0
    if options.race_start > 0 and options.title:
        if options.rider:
            outfile.write("\n\n\n\n\n\n\n\n\n %s\n {italic}%s{plain}{bold}\n" % (options.title, options.rider))
        else:
            outfile.write("\n\n\n\n\n\n\n\n\n\n %s\n" % (options.title,))
        start = timedelta(0, options.race_start, 0, 0)
        outfile.write(makeOffsetString(start))
        while times[startidx] <= start: startidx += 1
        
    endidx = len(times)-1
    
    if options.finale:
        endtime = duration - timedelta(0, 10)
        times.append(endtime)
        times.sort()
        endidx = times.index(endtime) - 1
        
    for idx in range(startidx,endidx):
        outfile.write("%s %.1f\n%s" % (textPrefix, 
            speeds[idx], makeOffsetString(times[idx+1])))
    
    if options.finale:
        outfile.write("%s %s\n" % (textPrefix, options.finale))
    else:
        outfile.write("%s %.1f\n" % (textPrefix, speeds[idx]))
    
    outfile.write(makeOffsetString(duration))
    
def cmdline():
    scriptname = os.path.basename(sys.argv[0]).split(".")[0].lower()
    usage=u"""usage: %prog [options] [infile [outfile]]
purpose: read a GPX file and produce a QuickTime text file to make subtitles with the speed in it. 

If infile and outfile are not given, stdin and stdout are assumed.

Copyright © 2008  Maarten Sneep

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version."""
    parser = OptionParser(usage=usage, version="%prog "+__version__+" ("+__date__+")")
    
    parser.add_option("", "--movie-start", dest="start",
        default=None, metavar="DATETIME", 
        help="The clip starts at DATETIME, in ISO format: '2008-08-09T15:28:45Z'.")
    parser.add_option("", "--movie-end", dest="end",
        default=None, metavar="DATETIME", 
        help="The clip ends at DATETIME, in ISO format: '2008-08-09T15:28:45Z'.")
    parser.add_option("", "--duration", dest="duration",
        default=None, metavar="DURATION",
        help="The length of the clip, in '[[hh:]mm:]ss'.")
    parser.add_option("", "--race-start", dest="race_start", 
        default=10, type="int", 
        metavar="START", 
        help="The actual start of the race, seconds into the clip. The title is shown up to this moment.")
    parser.add_option("", "--size", dest="size", default="0,0",
        metavar="WIDTH,HEIGHT",
        help="The size of the movie-frame in pixels.")
    parser.add_option("", "--location", dest="location", default="LB",
        metavar="LOC", 
        help="The location of the title, in two letters: T=top, B=bottom, L=left, R=right. Defaults to bottom-left.")
    
    parser.add_option("", "--title", dest="title", default="",
        metavar="'TITLE'", help="The TITLE is shown before the START.")
    parser.add_option("", "--rider", dest="rider", default="",
        metavar="'RIDER'", help="The name of the rider, used as subtitle.")
    parser.add_option("", "--finale", dest="finale", default="",
        metavar="'FINALE'", 
        help="Show the text FINALE during the last 10 seconds.")
    
    return parser.parse_args()

# Run this code when called directly instead of loaded as a module.
if __name__ == "__main__":
    Main()
